from typing import Any, Dict, Iterable, List, Optional, Tuple

from pydantic import Field, IPvAnyAddress, StrictStr, ConfigDict

from .schema import Domain, Schema, VersionStr, chainable


class Scanner(Schema):
    version: Optional[VersionStr] = Field(
        description="version of the microengine that generated the assertion",
        default = None,
    )
    polyswarmclient_version: Optional[VersionStr] = Field(description="version of polyswarmclient", default=None)
    vendor_version: Optional[str] = Field(description="version of the engine that generated the assertion", default=None)
    signatures_version: Optional[str] = Field(description="version of the engine's antimalware signatures", default=None)
    environment: Optional[Dict[str, Any]] = Field(description="analysis environment metadata", default=None)


class StixSignature(Schema):
    stix_schema: str = Field(alias='schema')
    signature: Any

    @property
    def schema(self):
        return self.stix_schema

    def dict(self, **kwargs):
        kwargs['by_alias'] = True
        return super().dict(**kwargs)


class ScanMetadata(Schema):
    model_config = ConfigDict(populate_by_name=True, extra='allow')

    malware_family: StrictStr = Field(
        default=...,
        description='name of the malware family specified by this microengine',
    )
    domains: Optional[List[Domain]] = []
    ip_addresses: Optional[List[IPvAnyAddress]] = []
    stix: Optional[List[StixSignature]] = []
    scanner: Optional[Scanner] = None
    heuristic: Optional[bool] = Field(description='indicator for assertions generated from heuristics', default=None)

    @property
    def extra(self):
        return [(k, v) for k, v in self.__dict__.items() if k not in self.model_fields]

    @chainable
    def add_domain(self, domain: Domain):
        self.domains.append(domain)

    @chainable
    def add_domains(self, domains: Iterable[Domain]):
        self.domains.extend(domains)

    @chainable
    def set_malware_family(self, malware_family: str):
        self.malware_family = malware_family

    @chainable
    def add_extra(self, key: str, value: Any):
        setattr(self, key, value)

    @chainable
    def add_extras(self, extras: Iterable[Tuple[str, Any]]):
        for k, v in extras:
            self.add_extra(k, v)

    @chainable
    def add_ip_address(self, ip_address: IPvAnyAddress):
        self.ip_addresses.append(ip_address)

    @chainable
    def add_ip_addresses(self, ip_addresses: Iterable[IPvAnyAddress]):
        self.ip_addresses.extend(ip_addresses)

    @chainable
    def add_stix_signature(self, stix_schema: str, signature: Any):
        self.stix.append(StixSignature(schema=stix_schema, signature=signature))

    @chainable
    def add_stix_signatures(self, signatures: Iterable[Tuple[str, Any]]):
        for signature in signatures:
            self.add_stix_signature(*signature)

    @chainable
    def set_analysis_conclusion(self, heuristic: bool = None):
        if heuristic is not None:
            self.heuristic = heuristic

    def set_scanner(self, **scanner):
        environment = {k: scanner.pop(k, None) for k in ('architecture', 'operating_system')}
        if any(environment.values()):
            scanner.setdefault('environment', {}).update(environment)
        self.scanner = Scanner(**scanner)
        return self


class Verdict(ScanMetadata):
    pass



from typing import List, Optional, Union

from pydantic import (
    AnyUrl,
    Field,
    PositiveInt,
    StrictStr,
    RootModel
)

from .schema import MD5, SHA1, SHA256, Schema


class FileArtifact(Schema):
    filename: Optional[str] = None
    filesize: Optional[PositiveInt] = None
    mimetype: StrictStr
    sha256: Optional[SHA256] = Field(title='SHA256', default=None)
    sha1: Optional[SHA1] = Field(title='SHA1', default=None)
    md5: Optional[MD5] = Field(title='MD5', default=None)


class URLArtifact(Schema):
    # protocol can actually be derived directly from `uri`, we keep it here to preserve a
    # preexisting interface.
    protocol: Optional[str]
    uri: AnyUrl = ...

    def __post_init__(self):
        if self.protocol is None:
            self.protocol = self.uri.protocol


class Bounty(RootModel, Schema):
    root: List[Union[FileArtifact, URLArtifact]] = Field(min_items=1, max_items=256, default=[])

    @property
    def artifacts(self):
        return self.root

    def add_file_artifact(self, **kwargs):
        self.root.append(FileArtifact(**kwargs))
        return self

    def add_url_artifact(self, uri: str = None, protocol: str = None):
        if uri is None:
            raise ValueError
        if protocol is not None:
            proto, *_ = protocol.rsplit('://', 1)
            uri = '{}://{}'.format(proto, next(reversed(uri.split('://', 1))))
        self.root.append(URLArtifact(uri=str(uri), protocol=protocol))
        return self

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]



import functools
import logging
from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Type,
    Union,
    cast,
)
from pydantic.fields import FieldInfo
from pydantic import BaseModel, ValidationError, constr, ConfigDict

logger = logging.getLogger(__name__)

MD5 = constr(pattern='^[0-9a-fA-F]{32}$', min_length=32, max_length=32)
SHA1 = constr(pattern='^[0-9a-fA-F]{40}$', min_length=40, max_length=40)
SHA256 = constr(pattern='^[0-9a-fA-F]{64}$', min_length=64, max_length=64)
Domain = constr(
    pattern=r'(?:{int_chunk}\.)*?{int_chunk}{int_domain_ending}'.format(
        int_chunk=r'[_0-9a-\U00040000](?:[-_0-9a-\U00040000]{0,61}[_0-9a-\U00040000])?',
        int_domain_ending=r'(?P<tld>(\.[^\W\d_]{2,63})|(\.(?:xn--)[_0-9a-z-]{2,63}))?\.?',
    ),
    min_length=3,
    max_length=61 + 63 + 3
)
VersionStr = constr(pattern=r"^[0-9]+([.][0-9]+)*$")


def chainable(fn: Callable):
    """
    Decorator to validate the arguments passed to a function, returning self
    """
    # @validate_arguments
    @functools.wraps(fn)
    def setter_wrapper(self, *args: Any, **kwargs: Any) -> Any:
        fn(self, *args, **kwargs)
        return self

    return setter_wrapper


class NoValidator:
    def validate_python(self, _, self_instance: BaseModel):
        object.__setattr__(self_instance, '__pydantic_fields_set__', set())
        object.__setattr__(self_instance, '__pydantic_extra__', {})
        object.__setattr__(self_instance, '__pydantic_private__', {})

        model_fields = cast(dict[str, FieldInfo], self_instance.model_fields)
        # print(model_fields)
        mapping = {}
        for field, info in model_fields.items():
            if not info.kw_only:
                if info.is_required():
                    mapping[field] = None
                else:
                    mapping[field] = info.get_default()
        for field, value in mapping.items():
            setattr(self_instance, field, value)
        return self_instance


class Schema(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    def __init__(self, *args, **kwargs):
        if args or kwargs:
            super().__init__(*args, **kwargs)
        else:
            with self.disable_validations():
                super().__init__()

    def __getitem__(self, k):
        try:
            return getattr(self, k)
        except AttributeError:
            raise KeyError

    @classmethod
    def get_schema(cls):
        """
        Get the path to the backing schema this Metadata object users
        :return: Tuple[string, string] where first string is the path,
        and the second is the schema name
        """
        return cls.model_json_schema()

    @contextmanager
    def disable_validations(self):
        validator = self.__pydantic_validator__
        try:
            object.__setattr__(self, '__pydantic_validator__', NoValidator())
            yield
        finally:
            object.__setattr__(self, '__pydantic_validator__', validator)

    def __str__(self):
        return self.json()

    def __eq__(self, other):
        if isinstance(other, dict):
            return self.dict(exclude_defaults=True) == other
        else:
            return super().__eq__(other)

    def dict(self, *args, by_alias=True, **kwargs):
        model_dict = super().model_dump(*args, **kwargs)
        super().model_validate(model_dict)
        return model_dict

    def json(self, *args, by_alias=True, **kwargs):
        kwargs.setdefault('exclude_defaults', True)
        # performing validation implicitly
        self.dict(*args, by_alias=by_alias, **kwargs)
        # use the custom json parser here
        return self.model_dump_json(*args, by_alias=by_alias, **kwargs)

    @classmethod
    def model_validate(cls: Type['BaseModel'], value: Any, **kwargs) -> 'Union[Schema, bool]':
        try:
            return BaseModel.model_validate.__func__(cls, value, **kwargs)
        except ValidationError as e:
            logger.error(e)
            return False
        return False

    @classmethod
    def _model_validate_old(cls: Type['BaseModel'], value: Any, **kwargs) -> 'Union[Schema, bool]':
        """
        Previous version of model_validate

        Kept to make easier to keep compatibility where needed
        """
        try:
            BaseModel.model_validate.__func__(cls, value)
        except ValidationError as e:
            logger.error(e)
            return False
        return True



from .schema import Schema
from .assertion import Assertion
from .bounty import Bounty, FileArtifact, URLArtifact
from .verdict import Verdict, Scanner, StixSignature, ScanMetadata

__all__ = [
    'Assertion',
    'Bounty',
    'Schema',
    'Verdict',
    'ScanMetadata',
    'FileArtifact',
    'URLArtifact',
    'Scanner',
    'StixSignature',
]



from typing import List

from pydantic import Field, RootModel

from .schema import Schema, chainable
from .verdict import Verdict


class Assertion(RootModel, Schema):
    root: List[Verdict] = Field(min_items=1, max_items=256, default=[])

    @property
    def artifacts(self):
        return self.root

    @chainable
    def add_artifact(self, verdict: Verdict):
        self.root.append(verdict)

    @chainable
    def add_artifacts(self, verdicts: List[Verdict]):
        self.root.extend(verdicts)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]
